import mock

import yaml
from yaxp import xpath as xp

mycommand = mock.Page(
    title="My Command",
    file="mycommand.md",
    ctype="markdown",
    content="""
# Help page

## Overview

- See <cmd:help2>

## Help

## Details
""",
    meta=yaml.safe_load("""
subtitle: page subtitle
terms:
  - cmd4
  - "cmd5#world2"
  - page: cmd2
  - page: "cmd3#world"
  - page:
    - help1
    - term2: hello
    - term3: details
  - cmd:
    - help2
    - help3: help
anchors:
  - world: Description of world
    """))

commands = mock.Page(
    title="Commands",
    file="commands.md",
    content="""
<body>

    <p>See <cmd:help2>

    <glossary::page|theme=detailed>

    <glossary::cmd>

    <glossary::_>
</body>


""")


def test_page_ref_simple(config):
    pages = mock.render([mycommand, commands], config)

    dl = xp.dl(_class="#mkdocs-ezglossary-summary", _id="page")
    dl = dl.has(xp.dt.a(href="../mycommand.md#",
                        text="help1"))
    assert len(pages['commands.md'].xpath(str(dl))) == 1

    dd = xp.dd.ul.li.a(href="../commands.md#cmd_help2_refs_0",
                       text="Commands")
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="cmd")
    dl = dl.has(dd)
    dl = dl.has(xp.dt.a(text="help2"))
    assert len(pages['commands.md'].xpath(str(dl))) == 1


def test_page_ref_anchor(config):
    pages = mock.render([mycommand, commands], config)

    a = xp.a(href="../mycommand.md#hello",
             text="term2")
    dl = xp.dl(_class="#mkdocs-ezglossary-summary", _id="page")
    dl = dl.has(xp.dt.has(a))
    assert len(pages['commands.md'].xpath(str(dl))) == 1


def test_page_ref_direct(config):
    pages = mock.render([mycommand, commands], config)

    dd = xp.dd
    a = xp.a(href="../mycommand.md#",
             text="cmd2")
    dd = dd.has(xp.p(text="*page subtitle"))
    dl = xp.dl(_class="#mkdocs-ezglossary-summary", _id="page")
    dl = dl.has(xp.dt.has(a))
    dl = dl.has(dd)
    assert len(pages['commands.md'].xpath(str(dl))) == 1


def test_page_ref_direct_anchor(config):
    pages = mock.render([mycommand, commands], config)

    a = xp.a(href="../mycommand.md#world",
             text="cmd3")
    dd = xp.dd()
    dd = dd.has(xp.p(text="*Description of world"))
    dl = xp.dl(_class="#mkdocs-ezglossary-summary", _id="page")
    dl = dl.has(xp.dt.has(a))
    dl = dl.has(dd)
    assert len(pages['commands.md'].xpath(str(dl))) == 1


def test_page_ref_default_section(config):
    pages = mock.render([mycommand, commands], config)

    a = xp.a(href="../mycommand.md#",
             text="cmd4")
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="_")
    dl = dl.has(xp.dt.has(a))
    assert len(pages['commands.md'].xpath(str(dl))) == 1


def test_page_ref_default_section_anchor(config):
    pages = mock.render([mycommand, commands], config)

    a = xp.a(href="../mycommand.md#world2",
             text="cmd5")
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="_")
    dl = dl.has(xp.dt.has(a))
    assert len(pages['commands.md'].xpath(str(dl))) == 1


def test_page_ref_link(config):
    pages = mock.render([mycommand, commands], config)

    dl = xp.body.p.a(_class="mkdocs-ezglossary-link",
                     name="cmd_help2_refs_0",
                     title="page subtitle",
                     href="../mycommand.md#",
                     text="help2")
    assert len(pages['commands.md'].xpath(str(dl))) == 1



import logging

from yaxp import xpath as xp

import mock

log = logging.getLogger(__name__)


def test_summary(simple, summary, config):
    config['inline_refs'] = "short"
    pages = mock.render([simple, summary], config)
    summary = pages["summary.md"]
    log.debug(summary)

    # first in test glossary
    a = xp.a(href="../simple.md#test_first_defs_0",
             text="first")
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="test")
    dl = dl.has(xp.dt.has(a))
    assert len(summary.xpath(str(dl))) == 1

    l1 = xp.a(href="../simple.md#test_first_refs_0")
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="test")
    dl = dl.has(xp.dl.has(a))
    assert len(summary.xpath(str(dl))) == 0

    a = xp.a(href="../simple.md#test_third_defs_0",
             text="third")
    dd = xp.dd
    l1 = xp.ul.li.a(href="../simple.md#test_third_refs_0",
                    text="Hello")
    l2 = xp.ul.li.a(href="../summary.md#test_third_refs_1",
                    text="Summary")
    dd = dd.has(l1)
    dd = dd.has(l2)
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="test")
    dl = dl.has(xp.dt.has(a))
    dl = dl.has(dd)
    assert len(summary.xpath(str(dl))) == 1

    a = xp.a(href="../simple.md#demo_first_defs_0",
             text="first")
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="demo")
    dl = dl.has(xp.dt.has(a))
    assert len(summary.xpath(str(dl))) == 1

    a = xp.a(href="../simple.md#demo_third_defs_0",
             text="third")
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="demo")
    dl = dl.has(xp.dt.has(a))
    assert len(summary.xpath(str(dl))) == 1

    a = xp.a(href="../simple.md#demo_third_refs_0",
             text="third")
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="demo")
    dl = dl.has(xp.dd.has(a))
    assert len(summary.xpath(str(dl))) == 0


def test_summary_noref(simple, summary, config):
    config['list_references'] = False
    pages = mock.render([simple, summary], config)
    summary = pages["summary.md"]

    a = xp.a(href="../simple.md#test_first_defs_0",
             text="first")
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="test")
    dl = dl.has(xp.dt.has(a))
    assert len(summary.xpath(str(dl))) == 1

    a = xp.a(href="../simple.md#test_third_refs_0",
             text="Hello")
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="test")
    dl = dl.has(xp.dt(text="third"))
    dl = dl.has(xp.dd.has(a))
    assert len(summary.xpath(str(dl))) == 0


def test_custom_summary(simple, summary, config):
    config['templates'] = "tests/custom"
    pages = mock.render([simple, summary], config)
    summary = pages["summary.md"]

    dd = xp.dd.ul.li.a(href="../simple.md#test_third_refs_0",
                       text="Hello")
    dl = xp.dl(_class="custom-summary", _id="test")
    dl = dl.has(xp.dt(text="third"))
    dl = dl.has(dd)
    assert len(summary.xpath(str(dl))) == 1


def test_default_summary(simple, summary, config):
    config['use_default'] = True
    pages = mock.render([simple, summary], config)
    summary = pages["summary.md"]
    log.debug(summary)

    a = xp.a(href="../simple.md#__default_defs_0",
             text="default")
    dl = xp.dl(_class="mkdocs-ezglossary-summary", _id="_")
    dl = dl.has(xp.dt.has(a))
    assert len(summary.xpath(str(dl))) == 1


def test_summary_table(simple, tablesummary, config):
    pages = mock.render([simple, tablesummary], config)
    summary = pages["tablesummary.md"]
    log.debug(summary)

    table = xp.div(_class="#mkdocs-ezglossary-summary").table(_id="demo")
    tr = xp.tr
    table = table.has(xp.thead.has(xp.th(text="Term"))
                      .has(xp.th(text="Definition"))
                      .has(xp.th(text="References")))
    tr = tr.has(xp.td(text="*first"))
    table.has(tr)
    assert len(summary.xpath(str(table))) == 1



import logging
import mock

from yaxp import xpath

log = logging.getLogger(__name__)


def test_link_no_title(simple, config):
    html = mock.render_single(simple, config)

    dl = xpath.body.p.a(name="test_third_refs_0",
                        title="",
                        href="../simple.md#test_third_defs_0",
                        text="third")
    assert len(html.xpath(str(dl))) == 1


def test_link_short_title(simple, config):
    config['tooltip'] = "short"
    html = mock.render_single(simple, config)
    log.debug(html)

    dl = xpath.body.p.a(name="test_third_refs_0",
                        title="third term",
                        href="../simple.md#test_third_defs_0",
                        text="third")
    assert len(html.xpath(str(dl))) == 1


def test_link_full_title(simple, config):
    config['tooltip'] = "full"
    html = mock.render_single(simple, config)
    log.debug(html)

    dl = xpath.body.p.a(_class="mkdocs-ezglossary-link",
                        name="test_third_refs_0",
                        title="*detailed description of third term",
                        href="../simple.md#test_third_defs_0",
                        text="third")
    assert len(html.xpath(str(dl))) == 1


def test_link_replace_html(simple, config):
    config['tooltip'] = "full"
    html = mock.render_single(simple, config)
    log.debug(html)

    dl = xpath.body.p.a(name="test_second_refs_0",
                        title="*this text is formatted",
                        href="../simple.md#test_second_defs_0",
                        text="second")
    assert len(html.xpath(str(dl))) == 1


def test_link_second_ref(simple, summary, config):
    config['tooltip'] = "full"
    summary = mock.render([simple, summary], config)['summary.md']
    log.debug(summary)

    dl = xpath.body.p.a(name="test_third_refs_1",
                        title="*third term",
                        href="../simple.md#test_third_defs_0",
                        text="third")
    assert len(summary.xpath(str(dl))) == 1


def test_link_default_ref_disabled(simple, summary, config):
    """ Ensure definitions for the default sections are
        ignored when the configuration `use_default` is set
        to `False`.
    """
    summary = mock.render([simple, summary], config)['summary.md']
    log.debug(summary)

    dl = xpath.body.p.a(_class="mkdocs-ezglossary-link",
                        name="__default_refs_0",
                        title="",
                        href="../simple.md#__default_defs_0",
                        text="default")
    assert len(summary.xpath(str(dl))) == 0


def test_link_default_ref_enabled(simple, summary, config):
    """ Ensure definitions for the default sections are
        replaced when the configuration `use_default` is set
        to `True`.
    """
    config['use_default'] = True
    summary = mock.render([simple, summary], config)['summary.md']
    log.debug(summary)

    dl = xpath.body.p.a(_class="mkdocs-ezglossary-link",
                        name="__default_refs_0",
                        title="",
                        href="../simple.md#__default_defs_0",
                        text="default")
    assert len(summary.xpath(str(dl))) == 1



import logging

log = logging.getLogger(__name__)


class Entry:
    """ An entry in the glossary. """

    def __init__(self, target, page, definition):
        self.target = target
        """ The anchor to directly point to this specific link. """

        self.page = page
        """ The url of the page to which this entry points to. """

        self.definition = definition
        """ The definition of the term. """

    def __repr__(self):
        return self.page.url


class Glossary:
    """ The complete glossary for all sections """

    def __init__(self):
        self.clear()

    def add(self, section, term, linktype, page, definition=None, anchor=None):
        log.debug(f"glossary.add({section}, {term}, {linktype}, '{definition}', {anchor})")
        links = self._links(section, term, linktype)
        _id = f"{section}_{term}_{linktype}_{len(links)}".replace(" ", "_")
        anchor = _id if anchor is None else anchor
        links[_id] = Entry(anchor, page, definition)
        return _id

    def has(self, section: str) -> bool:
        """ Check if the glossary has a section named **section**.

            Args:
                section:
                    The name of the section to check

            Returns:
                    True, if a section with the given name exists.
        """
        return section in self._glossary

    def get(self, section: str, term: str, linktype: str) -> list[Entry]:
        """ Get a list of [Entry][mkdocs_ezglossary_plugin.glossary.Entry] instances
            for a specific **term** in a **section**.

            Args:
                section:
                    The name of the section
                term:
                    The term for which the [Entry] instances should be retreived.
                linktype:
                    Defines which type of links should be returned.

                    `refs`
                    : returns all references to the term

                    `defs`
                    : returns all definitions for the term

            Returns:
                A list of either `definitions` or `references` for the given term.
        """
        links = self._links(section, term, linktype).values()
        return list(links)

    def definition(self, section: str, term: str) -> str:
        """ Get the definition for a term.

            Args:
                section:
                    The name of the section
                term:
                    The term to get the definition for

            Returns:
                The definition of the term
        """
        defs = self.get(section, term, 'defs')
        if not defs:
            return ""
        return defs[0].definition

    def terms(self, section: str) -> list[str]:
        """ Return a list of entries for a specific section.

            Args:
                section:
                    The name of the section.

            Returns:
                A list of all terms in the specific section.
        """
        keys = list(self._section(section).keys())
        keys.sort()
        return keys

    def clear(self):
        self._glossary = {}

    def _links(self, section, term, linktype):
        return self._term(section, term)[linktype]

    def _term(self, section, term):
        _section = self._section(section)
        if term not in _section:
            self._section(section)[term] = {'defs': {}, 'refs': {}}
        return self._section(section)[term]

    def _section(self, section):
        if section not in self._glossary:
            self._glossary[section] = {}
        return self._glossary[section]



import os

from jinja2 import Environment, PackageLoader, select_autoescape, FileSystemLoader

env = Environment(
    loader=PackageLoader("mkdocs_ezglossary_plugin"),
    autoescape=select_autoescape()
)


def render(file, config, **kwargs):
    tpl = load(file, config)
    return tpl.render(**kwargs)


def load(file, config):
    if not config.templates:
        return env.get_template(file)
    if not os.path.exists(os.path.join(config.templates, file)):
        return env.get_template(file)

    templateLoader = FileSystemLoader(searchpath=config.templates)
    templateEnv = Environment(loader=templateLoader)
    return templateEnv.get_template(file)



import logging
import mock

from yaxp import xpath


log = logging.getLogger(__name__)


def test_default_config(simple, config):
    html = mock.render_single(simple, config)
    log.debug(html)
    dl = xpath.dl
    dl = dl.has(xpath.dt.has(xpath.a(name="test_first_defs_0", text="first")))
    dl = dl.has(xpath.dd(text="first term"))
    dl = dl.has(xpath.dt.has(xpath.a(name="test_second_defs_0", text="second")))
    dl = dl.has(xpath.dd(text="*second term"))
    assert len(html.xpath(str(dl)))

    dl = xpath.dl
    dl = dl.has(xpath.dt.has(xpath.a(name="demo_first_defs_0", text="first")))
    dl = dl.has(xpath.dd(text="demo 1"))
    dl = dl.has(xpath.dt.has(xpath.a(name="demo_second_defs_0", text="second")))
    dl = dl.has(xpath.dd(text="demo 2"))
    assert len(html.xpath(str(dl)))

    dl = xpath.dl
    dl = dl.has(xpath.dd(text="*demo 2").has(xpath.a()))
    assert len(html.xpath(str(dl))) == 0


def test_inline_refs(simple, config):
    config['inline_refs'] = "short"
    html = mock.render_single(simple, config)

    dl = xpath.dl
    dl = dl.has(xpath.dt.has(xpath.a(name="test_first_defs_0", text="first")))
    dl = dl.has(xpath.dd(text="*first term").has(xpath.a()))
    assert len(html.xpath(str(dl))) == 0

    dl = xpath.dl()
    dl = dl.has(xpath.dt.has(xpath.a(name="test_third_defs_0", text="third")))
    dl = dl.has(xpath.dd(text="*third term").has(xpath.a(title="Hello",
                                                         href="../simple.md#test_third_refs_0",
                                                         text="*[1]")))
    assert len(html.xpath(str(dl))) == 1


def test_default_section(simple, config):
    config['use_default'] = True
    html = mock.render_single(simple, config)
    log.debug(html)
    dl = xpath.dl
    dl = dl.has(xpath.dt.has(xpath.a(name="__default_defs_0", text="default")))
    dl = dl.has(xpath.dd(text="default term"))
    assert len(html.xpath(str(dl)))

    dl = xpath.dl
    dl = dl.has(xpath.dt.has(xpath.a(name="demo_first_defs_0", text="first")))
    dl = dl.has(xpath.dd(text="demo 1"))
    dl = dl.has(xpath.dt.has(xpath.a(name="demo_second_defs_0", text="second")))
    dl = dl.has(xpath.dd(text="demo 2"))
    assert len(html.xpath(str(dl)))

    dl = xpath.dl
    dl = dl.has(xpath.dd(text="*demo 2").has(xpath.a()))
    assert len(html.xpath(str(dl))) == 0


def test_formatted_dt(simple, config):
    html = mock.render_single(simple, config)
    log.debug(html)
    dl = xpath.dl
    dl = dl.has(xpath.dt.bold.em.code.has(xpath.a(name="demo_formatted_defs_0", text="formatted")))
    dl = dl.has(xpath.dd(text="formatted dd"))
    assert len(html.xpath(str(dl)))



import logging
import re
import os
from html import parser

from mkdocs.plugins import BasePlugin, event_priority
from mkdocs import config
from mkdocs.config import config_options as co

from .glossary import Glossary

from . import template

log = logging.getLogger("mkdocs.plugins.ezglossary")


class __re:
    def __init__(self):
        self.ws = r"[\n ]*"
        self.section = r"(\w+)"
        self.term = r"([\w -]+)"
        self.text = r"([^>]+)"
        self.dt = rf"<dt>(<.*>)?{self.section}:{self.term}(<.*>)?<\/dt>"
        self.dt_default = rf"<dt>(<.*>)?{self.term}(<.*>)?<\/dt>"
        self.dd = r"<dd>\n?((.|\n)+?)<\/dd>"
        self.options = r"([\|\=\w\+]+)"


_re = __re()


class GlossaryConfig(config.base.Config):
    tooltip = config.config_options.Choice(('none', 'heading', 'full'), default="none")
    inline_refs = config.config_options.Choice(('none', 'short', 'full'), default="none")
    sections = co.ListOfItems(config.config_options.Type(str), default=[])
    section_config = co.ListOfItems(config.config_options.Type(dict), default=[])
    strict = config.config_options.Type(bool, default=False)
    list_references = config.config_options.Type(bool, default=True)
    list_definitions = config.config_options.Type(bool, default=True)
    use_default = config.config_options.Type(bool, default=False)
    templates = config.config_options.Type(str, default="")


class GlossaryPlugin(BasePlugin[GlossaryConfig]):
    def __init__(self):
        self._glossary = Glossary()
        self._uuid = "6251a85a-47d0-11ee-be56-0242ac120002"
        self._reflink = "6251a85a-47d0-11ee-be56-0242ac120002"

    def on_pre_build(self, config, **kwargs):
        if self.config.strict and "_" not in self.config.sections:
            self.config.sections.append("_")
        if self.config.strict and len(self.config.sections) == 0:
            log.error("ezglossary: no sections defined, but 'strict' is true, plugin disabled")
        self._glossary.clear()

    def on_page_markdown(self, content, page, config, files):
        attributes = page.meta

        def _get_definition(anchor):
            anchors = attributes.get('anchors')
            if anchors:
                for anchor_def in anchors:
                    if anchor in anchor_def:
                        return anchor_def[anchor]
            if 'subtitle' in attributes:
                return attributes['subtitle']
            return page.title

        def _add2section(section, term, anchor=None):
            if not anchor:
                if "#" in term:
                    (term, anchor) = term.split("#")
                else:
                    (term, anchor) = (term, "")
            definition = _get_definition(anchor)
            log.debug(f"add2section: {section}:{term}:{anchor} -> '{definition}'")
            self._glossary.add(section,
                               term,
                               'defs',
                               page,
                               definition,
                               anchor)

        ez = attributes.get('terms')
        log.debug(ez)
        if not ez:
            return content

        for entry in ez:
            if isinstance(entry, str):
                _add2section("_", entry)
                continue

            for (section, data) in entry.items():
                if isinstance(data, str):
                    _add2section(section, data)
                    continue

                for term in data:
                    if isinstance(term, str):
                        _add2section(section, term)
                    if isinstance(term, dict):
                        for term, anchor in term.items():
                            _add2section(section, term, anchor)
        return content

    @event_priority(5000)
    def on_page_content(self, content, page, config, files):
        content = self._find_definitions(content, page)
        content = self._register_glossary_links(content, page)
        return content

    def on_post_page(self, output, page, config):
        _dir = os.path.dirname(page.url)
        levels = len(_dir.split("/"))
        root = "../" * levels
        output = self._replace_glossary_links(output, page, root)
        output = self._replace_inline_refs(output, page, root)
        output = self._print_glossary(output, root)
        return output

    def _print_glossary(self, html, root):
        def _replace(mo):
            types = []
            section = mo.group(1)
            options = mo.group(2) or ""

            lr = "do_refs" in options
            if "no_refs" not in options and "do_refs" not in options:
                lr = self._get_config(section, 'list_references')
            if lr:
                types.append('refs')

            ld = "do_defs" in options
            if "no_defs" not in options and "do_defs" not in options:
                ld = self._get_config(section, 'list_definitions')
            if ld:
                types.append('defs')

            if len(types) == 0:
                log.warning("list_definitons and list_references disabled, summary will be empty")

            if not self._glossary.has(section=section):
                log.warning(f"no section '{section}' found in glossary")

            terms = self._glossary.terms(section)
            theme = ""
            for option in options.split("|"):
                if "theme" in option:
                    theme = "-" + option.split("=")[1]
            return template.render(f"summary{theme}.html",
                                   self.config,
                                   glossary=self._glossary,
                                   types=types,
                                   section=section,
                                   terms=terms,
                                   root=root)
            return html

        regex = rf"<glossary::{_re.section}\|?{_re.options}?>"
        return re.sub(regex, _replace, html)

    def _register_glossary_links(self, output, page):
        def _replace(mo):
            section = mo.group(1)
            section = "_" if section == "default" else section
            term = mo.group(2)
            text = mo.group(3)[1:] if mo.group(3) else term
            _id = self._glossary.add(section, term, 'refs', page)
            return f"{self._uuid}:{section}:{term}:<{text}>:{_id}"

        def _replace_default(mo):
            section = "_"
            term = mo.group(1)
            text = mo.group(2)[1:] if mo.group(2) else term
            _id = self._glossary.add(section, term, 'refs', page)
            return f"{self._uuid}:{section}:{term}:<{text}>:{_id}"

        regex = rf"<{_re.section}\:{_re.term}(\|{_re.text})?>"
        output = re.sub(regex, _replace, output)
        regex = rf"<{_re.term}\:(\|{_re.text})?>"
        return re.sub(regex, _replace_default, output)

    def _replace_inline_refs(self, output, page, root):
        def _replace(mo):
            section = mo.group(1)
            term = mo.group(2)

            mode = self._get_config(section, 'inline_refs')

            entries = self._glossary.get(section, term, 'refs')
            return template.render(f"refs-{mode}.html",
                                   self.config,
                                   entries=entries,
                                   root=root)

        regex = fr"{self._reflink}:{_re.section}:{_re.term}"
        return re.sub(regex, _replace, output)

    def _replace_glossary_links(self, output, page, root):
        def _replace(mo):
            section = mo.group(1)
            term = mo.group(2)
            text = mo.group(3)
            _id = mo.group(4)
            defs = self._glossary.get(section, term, 'defs')

            if len(defs) == 0:
                log.warning(f"page '{page.url}' referenes to undefined glossary entry {section}:{term}")
                return f'<a name="{_id}">{text}</a>'
            if len(defs) > 1:
                log.warning(f"multiple definitions found for <{section}:{term}>, linking to first:")
                for _def in defs:
                    log.warning(f"\t{_def}")
            entry = defs[0]
            entry.definition = _html2text(entry.definition)
            return template.render("link.html",
                                   root=root,
                                   config=self.config,
                                   entry=entry,
                                   text=text,
                                   target=_id)

        regex = fr"{self._uuid}:{_re.section}:{_re.term}:<{_re.text}>:(\w+)"
        return re.sub(regex, _replace, output)

    def _find_definitions(self, content, page):
        log.debug(f"_find_definitions({page})")

        def _add_entry(section, term, definition, fmt_pre, fmt_post):
            log.debug(f"found entry: {section}:{term}:{definition} {fmt_pre}:{fmt_post}")

            if self.config.tooltip == "none":
                _tooltip = ""
            if self.config.tooltip == "short":
                _tooltip = definition.split("\n")[0]
            if self.config.tooltip == "full":
                _tooltip = definition

            if section not in self.config.sections and self.config.strict:
                log.warning(f"ignoring undefined section '{section}' in glossary")
                return None

            _id = self._glossary.add(section, term, 'defs', page, _tooltip)

            inline_refs = self._get_config(section, 'inline_refs')
            reflink = f"\n{self._reflink}:{section}:{term}" if inline_refs != "none" else ""
            return template.render("definition.html",
                                   self.config,
                                   target=_id,
                                   term=term,
                                   definition=definition,
                                   reflink=reflink,
                                   fmt_pre=fmt_pre if fmt_pre else "",
                                   fmt_post=fmt_post if fmt_post else "")

        def _replace(mo):
            fmt_pre = mo.group(1)
            section = mo.group(2)
            term = mo.group(3)
            fmt_post = mo.group(4)
            definition = mo.group(5)
            rendered = _add_entry(section, term, definition, fmt_pre, fmt_post)
            return rendered if rendered else mo.group()

        def _replace_default(mo):
            section = "_"
            fmt_pre = mo.group(1)
            term = mo.group(2)
            fmt_post = mo.group(3)
            definition = mo.group(4)
            rendered = _add_entry(section, term, definition, fmt_pre, fmt_post)
            return rendered if rendered else mo.group()

        if self.config.use_default:
            regex_dt = re.compile(rf"{_re.dt_default}{_re.ws}{_re.dd}", re.MULTILINE)
            content = re.sub(regex_dt, _replace_default, content)

        regex_dt = re.compile(rf"{_re.dt}{_re.ws}{_re.dd}", re.MULTILINE)
        ret = re.sub(regex_dt, _replace, content)
        return ret

    def _get_section_config(self, section):
        for entry in self.config.section_config:
            if entry['name'] == section:
                return entry
        return None

    def _get_config(self, section, entry):
        cfg = self._get_section_config(section)
        if not cfg or entry not in cfg:
            ret = self.config[entry]
        else:
            ret = cfg[entry]
        log.debug(f"_get_config({section}, {entry}): {ret}")
        return ret


def _html2text(content):
    class HTMLFilter(parser.HTMLParser):
        def __init__(self):
            super().__init__()
            self.text = ""

        def handle_data(self, data):
            self.text += data

    f = HTMLFilter()
    log.debug(f"adding {content}")
    f.feed(content)
    return f.text.strip()
